class Route:
    BLOCK_USER = 'blockGJUser20'
    UNBLOCK_USER = 'unblockGJUser20'
    RATE_LEVEL_STARS = 'rateGJStars211'
    RATE_LEVEL_DEMON = 'rateGJDemon21'
    SUGGEST_LEVEL_STARS = 'suggestGJStars20'
    GET_USER_INFO = 'getGJUserInfo20'
    USER_SEARCH = 'getGJUsers20'
    GET_USER_TOP = 'getGJScores20'
    UPDATE_USER_SCORE = 'updateGJUserScore22'
    REPORT_LEVEL = 'reportGJLevel'
    DOWNLOAD_LEVEL = 'downloadGJLevel22'
    LEVEL_SEARCH = 'getGJLevels21'
    UPDATE_LEVEL_DESC = 'updateGJDesc20'
    GET_MAP_PACKS = 'getGJMapPacks21'
    GET_GAUNTLETS = 'getGJGauntlets21'
    DELETE_LEVEL = 'deleteGJLevelUser20'
    GET_SONG_INFO = 'getGJSongInfo'
    GET_FRIENDS = 'getGJUserList20'
    REMOVE_FRIEND = 'removeGJFriend20'
    GET_ACC_COMMENTS = 'getGJAccountComments20'
    GET_COMMENTS = 'getGJComments21'
    GET_PRIVATE_MESSAGES = 'getGJMessages20'
    READ_PRIVATE_MESSAGE = 'downloadGJMessage20'
    SEND_PRIVATE_MESSAGE = 'uploadGJMessage20'
    DELETE_PRIVATE_MESSAGE = 'deleteGJMessages20'
    GET_TIMELY = 'getGJDailyLevel'
    LOGIN = 'accounts/loginGJAccount'
    LOAD_DATA = 'accounts/syncGJAccountNew'
    CAPTCHA = 'accounts/captcha'
    MANAGE_ACCOUNT = 'accounts/accountManagement'
    CHANGE_PASSWORD = 'accounts/changepassword'
    CHANGE_USERNAME = 'accounts/changeusername'
    DELETE_ACC_COMMENT = 'deleteGJAccComment20'
    DELETE_LEVEL_COMMENT = 'deleteGJComment20'
    UPLOAD_COMMENT = 'uploadGJComment21'
    UPLOAD_ACC_COMMENT = 'uploadGJAccComment20'
    UPDATE_ACC_SETTINGS = 'updateGJAccSettings20'
    GET_FRIEND_REQUESTS = 'getGJFriendRequests20'
    READ_REQUEST = 'readGJFriendRequest20'
    DELETE_REQUEST = 'deleteGJFriendRequests20'
    ACCEPT_REQUEST = 'acceptGJFriendRequest20'
    SEND_REQUEST = 'uploadFriendRequest20'
    REQUEST_MODERATOR = 'requestUserAccess'
    GET_COMMENT_HISTORY = 'getGJCommentHistory'
    GET_ACCOUNT_URL = 'getAccountURL'
    LIKE_ITEM = 'likeGJItem211'
    NEWGROUNDS_SONG_LISTEN = 'https://www.newgrounds.com/audio/listen/'
